#
# COPYRIGHT
#
# All contributions by Thorsten Wagner:
# Copyright (c) 2017 - 2019, Thorsten Wagner.
# All rights reserved.
#
# ---------------------------------------------------------------------------
#         Do not reproduce or redistribute, in whole or in part.
#      Use of this code is permitted only under licence from Max Planck Society.
#            Contact us at thorsten.wagner@mpi-dortmund.mpg.de
# ---------------------------------------------------------------------------
#
# pylint: disable=C0330
import itertools
import numpy as np
from scipy.spatial import cKDTree
from .utils import Filament
from . import utils


class FilamentTracer:
    def __init__(
        self,
        boxes,
        orientation_image,
        filament_width,
        search_radius=60,
        angle_delta=10,
        min_number_boxes=2,
        rescale_factor=1,
        rescale_factor_x=1,
        rescale_factor_y=1,
        do_merging=True,
        box_distance=None,
        nosplit=False,
    ):
        """
        Traces filaments in the image using picked boxes as support points.

        :param orientation_image: Estimated orientation at each position in the image
        :param boxes: Picked boxes
        :param filament_width: Estimated filament width in pixel
        :param search_radius: Search radius to search for the next valid box.
        :param angle_delta: Allowed delta in orientation
        :param min_number_boxes: Minimum number of boxes a filament should contain
        :param rescale_factor: Resacle factor for box coordinates
        """
        self.filament_width = filament_width
        self.search_radius = search_radius
        self.angle_delta = angle_delta
        self.min_number_boxes = min_number_boxes
        self.rescale_factor = rescale_factor
        self.rescale_factor_x = rescale_factor_x
        self.rescale_factor_y = rescale_factor_y
        self.dir_delta = self.angle_delta * 1.75
        self.box_coords_data = None  # Init in update_box_coords_data
        self.box_coord_kdtree = None  # Init in update_box_coords_data
        self.boxes_distance_matrix = None  # Init in update_box_coords_data
        self.boxes = boxes
        self.orientation_image = orientation_image
        self.do_merging = do_merging
        self.box_distance = box_distance
        self.nosplit = nosplit

    @property
    def boxes(self):
        return self.__boxes

    @boxes.setter
    def boxes(self, boxes):

        self.__boxes = boxes
        self.update_box_coords_data()

    def update_box_coords_data(self):
        self.box_coords_data = np.empty(shape=(len(self.boxes), 2))
        for i, box in enumerate(self.boxes):
            box.x = box.x * self.rescale_factor
            box.y = box.y * self.rescale_factor
            box.w = box.w * self.rescale_factor
            box.h = box.h * self.rescale_factor
            box.info = i
            self.box_coords_data[i, 0] = box.x
            self.box_coords_data[i, 1] = box.y

        self.box_coord_kdtree = cKDTree(self.box_coords_data)
        self.boxes_distance_matrix = self.box_coord_kdtree.sparse_distance_matrix(
            self.box_coord_kdtree, 999999
        )

    def merge_filaments(self, filaments):
        search_radius_sq = self.search_radius * self.search_radius
        merged = True
        while merged:
            min_dist_filament = None
            min_dist_fil_a_index = None
            min_dist_fil_b_index = None
            merged = False
            for i in range(len(filaments)):
                for j in range((i + 1), len(filaments)):
                    if i != j:

                        if (
                            distancesq_filament(filaments[i], filaments[j])
                            > search_radius_sq
                        ):
                            # Skip if filaments are to far away
                            continue

                        merge_candidate = self.merge_filament(
                            filaments[i], filaments[j]
                        )
                        # print("Try to merge", i, j, merge_candidate is None)
                        if merge_candidate is not None:
                            if min_dist_filament is None:
                                min_dist_filament = merge_candidate
                                min_dist_fil_a_index = i
                                min_dist_fil_b_index = j
                            elif merge_candidate[1] < min_dist_filament[1]:
                                min_dist_filament = merge_candidate
                                min_dist_fil_a_index = i
                                min_dist_fil_b_index = j

            if min_dist_filament is not None:
                if min_dist_fil_a_index > min_dist_fil_b_index:
                    del filaments[min_dist_fil_a_index]
                    del filaments[min_dist_fil_b_index]
                else:
                    del filaments[min_dist_fil_b_index]
                    del filaments[min_dist_fil_a_index]
                filaments.append(min_dist_filament[0])
                merged = True

        return filaments

    def trace_filaments(self):
        """
        Traces filaments based on boxes provided by crYOLO + orientation image
        :return: List of traced filements
        """
        assigned_boxes = []
        filaments = []
        # while len(assigned_boxes) < len(self.boxes):
        for box in self.boxes:

            if not contains_box(box, assigned_boxes):  # box not in assigned_boxes:
                fil = Filament()

                fil.add_box(box)

                assigned_boxes.append(box)

                new_box_found = True

                while new_box_found:
                    new_box_found = False
                    candidate_boxes = self.get_box_neightborhood(
                        fil.boxes[-1], self.search_radius
                    )
                    candidate_boxes = [
                        box
                        for box in candidate_boxes
                        if not contains_box(box, assigned_boxes)
                    ]
                    next_box_indices, ignore = self.next_valid_box(
                        filament=fil,
                        ref_box=fil.boxes[-1],
                        candidate_boxes=candidate_boxes,
                    )

                    for i in next_box_indices:
                        new_box_found = True

                        fil.add_box(candidate_boxes[i])
                        assigned_boxes.append(candidate_boxes[i])
                    for i in ignore:
                        assigned_boxes.append(candidate_boxes[i])
                filaments.append(fil)
        #
        # Merge filaments
        #
        if self.do_merging:
            filaments = self.merge_filaments(filaments)

        filaments = [f for f in filaments if f.get_num_boxes() >= self.min_number_boxes]
        # Smooth
        for fil in filaments:
            fil.boxes = moving_window(boxes=fil.boxes, window_width=3)
        # Rescale
        for fil in filaments:
            for box in fil.boxes:
                box.x = box.x / self.rescale_factor
                box.y = box.y / self.rescale_factor
                box.w = box.w / self.rescale_factor
                box.h = box.h / self.rescale_factor

        # Resample
        filaments = utils.resample_filaments(filaments, self.box_distance)

        # Make straight filaments
        if not self.nosplit:
            filaments = split_filaments_by_straightness(filaments)

        # NMS for filaments:
        filaments = nms_for_filaments(
            filaments, self.filament_width / self.rescale_factor
        )
        return filaments

    def merge_filament(self, filament_a, filament_b):
        filament_a = Filament(filament_a.boxes.copy())
        filament_b = Filament(filament_b.boxes.copy())
        a_start = filament_a.boxes[0]
        a_end = filament_a.boxes[-1]
        b_start = filament_b.boxes[0]
        b_end = filament_b.boxes[-1]

        boxes = [b_start, b_end]

        candidate_box_1_index, _ = self.next_valid_box(
            filament=None, ref_box=a_start, candidate_boxes=boxes, debug=False
        )

        candidate_box_1 = None
        if len(candidate_box_1_index) > 0:
            candidate_box_1 = boxes[candidate_box_1_index[0]]
        boxes = [b_start, b_end]
        candidate_box_2_index, _ = self.next_valid_box(
            filament=None, ref_box=a_end, candidate_boxes=boxes, debug=True
        )
        candidate_box_2 = None
        if len(candidate_box_2_index) > 0:
            candidate_box_2 = boxes[candidate_box_2_index[0]]

        result_tuble = None
        if candidate_box_1 is not None:
            if box_are_equal(candidate_box_1, b_start):
                # Connects a_start with b_start
                reversed_b = list(reversed(filament_b.boxes))
                is_compatible_dir = merge_is_compatible_direction(
                    basis_filament=filament_a,
                    to_add_filament=Filament(reversed_b),
                    from_head_a=True,
                )
                if is_compatible_dir:
                    reversed_b.extend(filament_a.boxes)
                    new_filament_boxes = reversed_b
                    new_filament = Filament()
                    new_filament.boxes = new_filament_boxes
                    sq_dist = self.get_distance(a_start, b_start)
                    result_tuble = (new_filament, sq_dist)

            elif box_are_equal(candidate_box_1, b_end):
                # TODO: Do the same for the otehr cases...
                filament_b_for_dir_est = filament_b
                if len(filament_b.boxes) < 3:
                    filament_b_for_dir_est = Filament(filament_b.boxes)
                    for b in filament_a.boxes[:-2]:
                        filament_b_for_dir_est.boxes.append(b)

                is_compatible_dir = merge_is_compatible_direction(
                    basis_filament=filament_a,
                    to_add_filament=filament_b_for_dir_est,
                    from_head_a=True,
                )
                if is_compatible_dir:
                    new_filament_boxes = list(filament_b.boxes)
                    new_filament_boxes.extend(filament_a.boxes)
                    new_filament = Filament()
                    new_filament.boxes = new_filament_boxes
                    sq_dist = self.get_distance(a_start, b_end)
                    result_tuble = (new_filament, sq_dist)

        elif candidate_box_2 is not None:
            if box_are_equal(candidate_box_2, b_start):
                is_compatible_dir = merge_is_compatible_direction(
                    basis_filament=filament_a,
                    to_add_filament=filament_b,
                    from_head_b=True,
                )
                if is_compatible_dir:
                    new_filament_boxes = list(filament_a.boxes)
                    new_filament_boxes.extend(filament_b.boxes)

                    new_filament = Filament()
                    new_filament.boxes = new_filament_boxes
                    sq_dist = self.get_distance(a_end, b_start)
                    result_tuble = (new_filament, sq_dist)

            elif box_are_equal(candidate_box_2, b_end):
                reversed_b = list(reversed(filament_b.boxes))

                is_compatible_dir = merge_is_compatible_direction(
                    basis_filament=filament_a,
                    to_add_filament=Filament(reversed_b),
                    from_head_b=True,
                )
                if is_compatible_dir:

                    new_filament_boxes = list(filament_a.boxes)
                    new_filament_boxes.extend(reversed_b)

                    new_filament = Filament()
                    new_filament.boxes = new_filament_boxes
                    sq_dist = self.get_distance(a_end, b_end)

                    result_tuble = (new_filament, sq_dist)
        return result_tuble


    def remove_parallel(self, filaments):
        """
        Remove parallel lines
        :param filaments: List of filaments
        :param box_coords_data: Coordinates of boxes as numpy array
        :param filament_width: Estimated widht of filaments
        :return: None
        """
        new_column = np.ones((len(self.box_coords_data), 1))
        new_column = new_column * -1
        box_coords_data = np.append(self.box_coords_data, new_column, axis=1)
        box_coords_data = np.append(box_coords_data, new_column, axis=1)

        for fil_index, fil in enumerate(filaments):
            len_fil = len(fil.boxes)
            for box in fil.boxes:
                box_coords_data[box.info, 2] = fil_index
                box_coords_data[box.info, 3] = len_fil

        for fil in filaments:
            remove_boxes = []
            for box in fil.boxes:
                boxes_indicies_in_radius = self.box_coord_kdtree.query_ball_point(
                    [box.x, box.y], self.filament_width / 2.0
                )
                boxes_of_filament_to_remove = [
                    index
                    for index in boxes_indicies_in_radius
                    if box_coords_data[index, 2] != -1
                    and box_coords_data[index, 2]
                    != box_coords_data[
                        box.info, 2
                    ]  # should belong to a different filament
                    and box_coords_data[box.info, 3] < box_coords_data[index, 3]
                ]  # should be shorter

                if boxes_of_filament_to_remove:
                    remove_boxes.append(box)
            fil.boxes = [box for box in fil.boxes if box not in remove_boxes]

    def get_box_neightborhood(self, ref_box, search_radius):
        """
        Finds neighboring boxes in a specific radius
        :param ref_box:
        :param search_radius:
        :return: List of box neighbors
        """
        box_indicies_in_search_radius = self.box_coord_kdtree.query_ball_point(
            [ref_box.x, ref_box.y], search_radius
        )
        boxes_in_search_radius = [
            self.boxes[index] for index in box_indicies_in_search_radius
        ]
        return boxes_in_search_radius

    def next_valid_box(self, filament, candidate_boxes, ref_box, debug=False):
        """
        Find the next valid box for a given filament
        :param filament: All boxes taht belong to the filament.
        Mainly used for orientation estimation
        :param ref_box: Refrence box. Outgoing from this the next valid box is searched.
        :param assigned_boxes: Boxes that are already assigned
        :param candidate_boxes: Candidate boxes

        :return: Next valid box
        """

        if filament is None:
            ref_orientation = get_orientation_for_box(ref_box, self.orientation_image)
        else:
            ref_orientation = filament_orientation(filament, self.orientation_image)

        ref_tracing_direction = filament_direction(filament)

        # Get boxes inside angular search region
        # ref_box_ = ref_box
        # if filament is not None:
        #    xavg = np.average([b.x for b in filament.boxes[-5:]])
        #    yavg = np.average([b.y for b in filament.boxes[-5:]])
        #    ref_box_ = utils.BoundBox(xavg,yavg,0,0)
        selected_candidate_boxes = [
            box
            for box in candidate_boxes
            if in_angular_search_region(
                ref_box=ref_box,
                candidate_box=box,
                angle_delta=self.dir_delta,
                ref_angle=ref_orientation,
            )
        ]

        to_ignore = [
            box
            for box in candidate_boxes
            if not in_angular_search_region(
                ref_box=ref_box,
                candidate_box=box,
                angle_delta=self.dir_delta,
                ref_angle=ref_orientation,
            )
        ]

        to_ignore = [
            box
            for box in to_ignore
            if self.get_distance(ref_box, box) < np.sqrt(self.filament_width ** 2) / 2
        ]

        # Check if in search radius
        selected_candidate_boxes = [
            box
            for box in selected_candidate_boxes
            if self.get_distance(ref_box, box) < self.search_radius
        ]

        # Check if in search range
        selected_candidate_boxes = [
            box
            for box in selected_candidate_boxes
            if angle_between_two_boxes(ref_box, box) < self.angle_delta
        ]

        # Check if same orientation

        selected_candidate_boxes = [
            box
            for box in selected_candidate_boxes
            if is_valid(
                ref_orientation,
                self.angle_delta,
                get_orientation_for_box(box, self.orientation_image),
            )
        ]

        if ref_tracing_direction is None and filament is not None:
            max_coh = 0
            max_coh_dir = None
            for bo_i, bo in enumerate(selected_candidate_boxes):
                help_dir = box_diff_direction_vector(bo, filament.boxes[-1])
                coherrence = 0
                for bo_j, bo2 in enumerate(selected_candidate_boxes):
                    if bo_i != bo_j:
                        help2_dir = box_diff_direction_vector(bo2, filament.boxes[-1])
                        if angle_between_two_vec(help_dir, help2_dir) < self.dir_delta:
                            coherrence = coherrence + 1
                if coherrence > max_coh:
                    max_coh_dir = help_dir
                    max_coh = max_coh
            ref_tracing_direction = max_coh_dir

        # If ref_tracing_direction, check same direction
        # Select candidates according direction ref_tracing_direction
        if ref_tracing_direction is not None and filament is not None:

            selected_candidate_boxes = [
                box
                for box_i, box in enumerate(selected_candidate_boxes)
                if angle_between_two_vec(
                    ref_tracing_direction,
                    # filament_direction(Filament(filament.boxes[-4:]).add_box(box)),
                    box_diff_direction_vector(box, filament.boxes[-1]),
                )
                < self.dir_delta
            ]

        nearest_valid_boxes_indices = [
            candidate_boxes.index(box) for box in selected_candidate_boxes
        ]

        to_delete_indices = [candidate_boxes.index(box) for box in to_ignore]

        nearest_valid_box_distancesq = [
            self.get_distance(ref_box, box) for box in selected_candidate_boxes
        ]

        # Sort according distances
        sorted_indeces = np.argsort(nearest_valid_box_distancesq)
        nearest_valid_boxes_indices = [
            nearest_valid_boxes_indices[i] for i in sorted_indeces
        ]

        return nearest_valid_boxes_indices, to_delete_indices

    def get_distance(self, boxa, boxb):
        """
        Calculates the euclidian distance between two boxes
        :param boxa: First box
        :param boxb: Second box
        :return: Euclidian distance between boxa and boxb
        """
        if (
            self.boxes_distance_matrix is not None
            and boxa.info is not None
            and boxb.info is not None
        ):
            return self.boxes_distance_matrix[boxa.info, boxb.info]
        return np.sqrt(sq_distance_between_two_boxes(boxa, boxb))


def merge_is_compatible_direction(
    basis_filament,
    to_add_filament,
    filament_direction_delta=50,
    from_head_a=False,
    from_head_b=False,
):
    if len(basis_filament.boxes) < 2:
        boxes = [basis_filament.boxes[0], to_add_filament.boxes[0]]
        a_filament = Filament(boxes)
    else:
        a_filament = basis_filament

    if len(to_add_filament.boxes) < 2:
        boxes = [
            basis_filament.boxes[len(basis_filament.boxes) - 1],
            to_add_filament.boxes[0],
        ]
        b_filament = Filament(boxes)
    else:
        b_filament = to_add_filament

    dira = filament_direction(a_filament, from_head=from_head_a)

    dirb = filament_direction(b_filament, from_head=from_head_b)

    angle_box_dir = angle_between_two_vec(dira, dirb)
    return angle_box_dir < filament_direction_delta


def nms_for_filaments(filaments, filament_width):
    # NMS for filaments
    filaments_nms_remove_indices = []
    new_filaments = []
    for i in range(len(filaments)):
        for j in range((i + 1), len(filaments)):
            if i != j:
                if not is_in_search_radius(filaments[i], filaments[j], filament_width):
                    continue
                long = i
                short = j
                if len(filaments[i].boxes) < len(filaments[j].boxes):
                    long = j
                    short = i

                split_indices = []
                for box_combination_indices in itertools.product(
                    range(len(filaments[short].boxes)),
                    range(len(filaments[long].boxes)),
                ):
                    dist = sq_distance_between_two_boxes(
                        filaments[short].boxes[box_combination_indices[0]],
                        filaments[long].boxes[box_combination_indices[1]],
                    )
                    if dist < filament_width ** 2 / 4:
                        if box_combination_indices[0] not in split_indices:
                            split_indices.append(box_combination_indices[0])
                if split_indices:
                    if short not in filaments_nms_remove_indices:
                        filaments_nms_remove_indices.append(short)
                    if len(filaments[short].boxes) > 1:
                        splitted_filament = split_filamant(
                            filaments[short], split_indices
                        )
                        new_filaments.extend(splitted_filament)

    filaments_nms_remove = [
        filaments[fil_index] for fil_index in filaments_nms_remove_indices
    ]
    for fil in filaments_nms_remove:
        filaments.remove(fil)
    for fil in new_filaments:
        filaments.append(fil)

    return filaments


def is_in_search_radius(filament_a, filament_b, search_radius):
    """
    If some boxes of two filaments have a distance smaller then the given radius
    :param filament_a: First filament
    :param filament_b: Second filament
    :param search_radius: Search radius
    :return: True if at least one box pair of the two filaments have a distance smaller then the given radius
    """
    search_radius_sq = search_radius * search_radius
    for box_combination in itertools.product(filament_a.boxes, filament_b.boxes):
        dist = sq_distance_between_two_boxes(box_combination[0], box_combination[1])
        if dist < search_radius_sq:
            return True
    return False


def split_filamant(filament, split_indices):
    """
    Splits a filament at specified inidices
    :param filament: Filament to be splitted
    :param split_indices: list of indicies where the filament is splitted
    :return: List of filament segments
    """
    segments = []
    split_indices.sort()

    boxes_new_a = filament.boxes[(split_indices[-1] + 1) :]
    if boxes_new_a:
        new_fil_a = Filament(boxes_new_a)
        segments.append(new_fil_a)

    if split_indices[0] > 0:
        boxes_new_b = filament.boxes[: split_indices[0]]
        new_fil_b = Filament(boxes_new_b)
        segments.append(new_fil_b)

    if len(split_indices) > 1:
        for i in range(len(split_indices) - 1):
            boxes_new = filament.boxes[(split_indices[i] + 1) : split_indices[i + 1]]
            if boxes_new:
                new_fil = Filament(boxes_new)
                segments.append(new_fil)
    return segments


def moving_window(boxes, window_width=3):
    """
    Applied a moving average filter on a set of boxes
    :param boxes: The boxes
    :param window_width: Width width for filtering
    :return: Smoothed set of boxes
    """
    if len(boxes) < (window_width + 1):
        return boxes

    new_boxes = []
    offset = int((window_width - 1) / 2)
    for i in range(offset):
        new_boxes.append(boxes[i])
    for i in range(offset, len(boxes) - offset):
        mean_x = 0
        mean_y = 0
        mean_w = 0
        mean_h = 0
        mean_c = 0
        window_range = range(i - offset, i + offset + 1)
        for j in window_range:
            mean_x += boxes[j].x
            mean_y += boxes[j].y
            mean_w += boxes[j].w
            mean_h += boxes[j].h
            if boxes[j].c is None:
                mean_c += 1
            else:
                mean_c += boxes[j].c
        mean_x = int(mean_x / len(window_range))
        mean_y = int(mean_y / len(window_range))
        mean_w = int(mean_w / len(window_range))
        mean_h = int(mean_h / len(window_range))
        mean_c = mean_c / len(window_range)
        new_box = utils.BoundBox(x=mean_x, y=mean_y, w=mean_w, h=mean_h, c=mean_c)
        new_box.info = i
        new_boxes.append(new_box)
    for i in range(len(boxes) - offset, len(boxes)):
        new_boxes.append(boxes[i])
    return new_boxes


def box_are_equal(box1, box2):
    """
    :param box1: First box
    :param box2: Second box
    :return: True, if x,y, width and height are equal
    """
    if (
        int(box1.x) == int(box2.x)
        and int(box1.y) == int(box2.y)
        and int(box1.w) == int(box2.w)
        and int(box1.h) == int(box2.h)
    ):
        return True
    return False


def filter_filaments_by_num_boxes(filaments, minimum_number_boxes):
    """
    Removes filaments that are shorter than a minimum number of boxes.
    :param filaments: List of filaments
    :param minimum_number_boxes: Minimum number of boxes per filament
    :return: All filaments that have more boxes than minimum_number_boxes
    """
    filtered_filaments = [f for f in filaments if len(f.boxes) > minimum_number_boxes]
    return filtered_filaments


def split_filaments_by_straightness(resamples_filaments, straightness_threshold=0.95):
    """
    Splits multiple filaments by straightness. See split_filament_by_straightness.
    :param resamples_filaments: List of filaments
    :param straightness_threshold: straightness threshold.
    :return: List of splitted filaments
    """
    filamants_which_were_splitted = []
    new_fiament_segments = []
    for fil in resamples_filaments:
        segments = split_filament_by_straightness(fil, straightness_threshold)

        if len(segments) > 1:
            filamants_which_were_splitted.append(fil)
            new_fiament_segments.extend(segments)

    for fil in filamants_which_were_splitted:
        resamples_filaments.remove(fil)

    if new_fiament_segments:
        resamples_filaments.extend(new_fiament_segments)
    return resamples_filaments


def split_filament_by_straightness(filament, straightness_threshold=0.95):
    """
    Splits a filament into multiple subfilaments based on its straightness
    :param filament:
    :param straightness_threshold:
    :return: List of splitted sub-filaments
    """

    if _get_straightness(filament.boxes) >= straightness_threshold:
        return [filament]

    list_splitted_filaments = [filament]
    last_list_size = 0

    while len(list_splitted_filaments) > last_list_size:

        last_list_size = len(list_splitted_filaments)
        for fil in list_splitted_filaments:
            if _get_straightness(fil.boxes) >= straightness_threshold:
                continue

            max_straightness = -1
            max_straightness_index = -1
            for i in range(2, len(fil.boxes) - 2):
                straightness_to_i = _get_straightness(fil.boxes[:i])
                straightness_from_i = _get_straightness(fil.boxes[i:])
                mean_straightness = (straightness_to_i + straightness_from_i) / 2
                if mean_straightness > max_straightness:
                    max_straightness = mean_straightness
                    max_straightness_index = i

            if max_straightness_index > -1:
                boxes_new_a = fil.boxes[max_straightness_index:]
                boxes_new_b = fil.boxes[:max_straightness_index]
                new_fil_a = Filament(boxes_new_a)
                new_fil_b = Filament(boxes_new_b)
                list_splitted_filaments.remove(fil)
                list_splitted_filaments.append(new_fil_a)
                list_splitted_filaments.append(new_fil_b)
    return list_splitted_filaments


def _get_straightness(boxes):
    """
    Calculates the straightness for a set of boxes
    :param boxes: The boxes
    :return: Straightness value
    """
    len_sum = 0
    for i in range(1, len(boxes)):
        len_sum += np.sqrt(
            (boxes[i].x - boxes[i - 1].x) ** 2 + (boxes[i].y - boxes[i - 1].y) ** 2
        )
    len_start_end = np.sqrt(
        (boxes[0].x - boxes[len(boxes) - 1].x) ** 2
        + (boxes[0].y - boxes[len(boxes) - 1].y) ** 2
    )

    return len_start_end / len_sum


def distance_perpendicluar_to_box_direction(ref_box, candiate_box, ref_angle):

    ref_angle_rad = ref_angle / 180.0 * np.pi
    p1x = ref_box.x
    p1y = ref_box.y
    p2x = ref_box.x + np.cos(ref_angle_rad)
    p2y = ref_box.y + np.sin(ref_angle_rad)
    dist = np.abs(
        (p2y - p1y) * candiate_box.x
        - (p2x - p1x) * candiate_box.y
        + p2x * p1y
        - p2y * p1x
    ) / np.sqrt((p2y - p1y) ** 2 + (p2x - p1x) ** 2)
    return dist


def in_angular_search_region(ref_box, candidate_box, ref_angle, angle_delta):
    """

    :param ref_box:
    :param candidate_box:
    :param ref_angle: orientation in degree
    :param angle_delta: max angle delta in degree
    :return:
    """

    ref_angle_rad = ref_angle / 180.0 * np.pi
    ref_angle_vector = np.asarray([np.cos(ref_angle_rad), np.sin(ref_angle_rad)])
    ref_angle_vector_180 = -1 * ref_angle_vector

    cand_angle_vector = (candidate_box.x - ref_box.x, candidate_box.y - ref_box.y)
    norm = np.linalg.norm(cand_angle_vector)
    cand_angle_vector = cand_angle_vector / norm

    delta = angle_between_two_vec(ref_angle_vector, cand_angle_vector)
    delta_180 = angle_between_two_vec(ref_angle_vector_180, cand_angle_vector)
    in_search_region = delta < angle_delta or delta_180 < angle_delta
    return in_search_region


def get_orientation_for_box(box, orientation_image):
    index_y = int(box.y + box.h / 2)
    index_x = int(box.x + box.w / 2)
    index_y = (
        index_y
        if index_y < orientation_image.shape[0]
        else orientation_image.shape[0] - 1
    )
    index_x = (
        index_x
        if index_x < orientation_image.shape[1]
        else orientation_image.shape[1] - 1
    )
    return orientation_image[index_y, index_x]


def filament_orientation(filament, orientation_image, len_window=5):
    if filament is None:
        return None
    if len(filament.boxes) < 2:
        return get_orientation_for_box(filament.boxes[0], orientation_image)
    mean_orientation = 0
    last5_boxes = filament.boxes[-len_window:]
    for i in range(1, len(last5_boxes)):
        mean_orientation += get_orientation_for_box(last5_boxes[i], orientation_image)
    ref_orientation = mean_orientation / (len(last5_boxes) - 1)

    return ref_orientation


def filament_direction(filament, num_last_boxes=5, from_head=False):
    """
    Estimates the current filament direction (vector) based on the last num_last_positions boxes.
    :param filament: Filament
    :param num_last_boxes: Number of boxes to estimates the current directions
    :param from_head: Direction of estimation
    :return: Current direction as a vector
    """
    if filament is None:
        return None
    if len(filament.boxes) < 2:
        return None
    mean_y = 0
    mean_x = 0
    if from_head:
        last5_boxes = filament.boxes[:num_last_boxes]
    else:
        last5_boxes = filament.boxes[-num_last_boxes:]

    for i in range(1, len(last5_boxes)):
        vec_box_normed = box_diff_direction_vector(last5_boxes[i], last5_boxes[i - 1])
        mean_x += vec_box_normed[0]
        mean_y += vec_box_normed[1]
    ref_vec = (mean_x / (len(last5_boxes) - 1), mean_y / (len(last5_boxes) - 1))

    return ref_vec


def box_diff_direction_vector(box_a, box_b):
    norm_vec = (box_a.x - box_b.x, box_a.y - box_b.y)
    norm_vec = norm_vec / np.linalg.norm(norm_vec)
    return norm_vec


def sq_distance_between_two_boxes(box_a, box_b):
    """
    Calculates the squared euclidian distance between two boxes
    :param box_a: First box
    :param box_b: Second box
    :return: Squared distance between two boxes
    """
    return (box_a.x - box_b.x) ** 2 + (box_a.y - box_b.y) ** 2


def angle_between_two_boxes(box_a, box_b):
    """
    Calculates the angle (degree) between the centers of two boxes
    :param box_a: First box
    :param box_b: Second box
    :return: Angle between the centers of two boxes (degree)
    """
    vec_box_a = (box_a.x, box_a.y)
    vec_box_a = vec_box_a / np.linalg.norm(vec_box_a)
    vec_box_b = (box_b.x, box_b.y)
    vec_box_b = vec_box_b / np.linalg.norm(vec_box_b)

    return angle_between_two_vec(vec_box_a, vec_box_b)


def angle_between_two_vec(vec_a, vec_b):
    """
    Calculates the angle (degree) between the centers of two vectors

    :param vec_a: First vector
    :param vec_b: Second vector
    :return: Angle between the centers of two boxes (degree)
    """
    angle_rad = np.arccos(np.clip(np.dot(vec_a, vec_b), -1.0, 1.0))
    angle = angle_rad / (2 * np.pi) * 360
    return angle


def is_valid(ref_angle, delta_angle, test_angle, debug=False):
    """
    Checks if the differencen between two angles (in degree) is less then a give delta_angle
    :param ref_angle: Reference angleis_valid
    :param delta_angle: Allowed delta
    :param test_angle: Test angle
    :return: True if the difference is lower than the allowed delta
    """

    delta_1 = np.abs(ref_angle - test_angle)
    ref_2 = np.abs(ref_angle - 180)
    delta_2 = np.abs(ref_2 - test_angle)

    is_angle_valid = delta_1 < delta_angle or delta_2 < delta_angle
    if debug:
        print("V", delta_1, delta_2)
    return is_angle_valid


def contains_box(box, list_box):
    contains = False
    for b in list_box:
        if int(box.x) == int(b.x) and int(box.y) == int(b.y):
            return True

    return contains


def distancesq_filament(filament_a, filament_b):
    """

    :param filament_a: First filament
    :param filament_b: Second filament
    :return: Minimum distance between start and end points of the two filaments
    """
    start_end_boxes_a = [filament_a.boxes[0], filament_a.boxes[-1]]
    start_end_boxes_b = [filament_b.boxes[0], filament_b.boxes[-1]]
    min_dist = 999999999
    for start_end_combination in itertools.product(
        start_end_boxes_a, start_end_boxes_b
    ):
        dist = sq_distance_between_two_boxes(
            start_end_combination[0], start_end_combination[1]
        )
        if dist < min_dist:
            min_dist = dist
    return min_dist



#
# COPYRIGHT
#
# All contributions by Thorsten Wagner:
# Copyright (c) 2017 - 2019, Thorsten Wagner.
# All rights reserved.
#
# ---------------------------------------------------------------------------
#         Do not reproduce or redistribute, in whole or in part.
#      Use of this code is permitted only under licence from Max Planck Society.
#            Contact us at thorsten.wagner@mpi-dortmund.mpg.de
# ---------------------------------------------------------------------------
#
import multiprocessing
import os
import warnings

import numpy as np
import scipy.fftpack as fft

import cryolo.imagereader as imagereader

LOCK = None


def window(image, width):
    """

    :param image:
    :param width: Width of the gaussian filter in fourier pixel
    :return:
    """
    center_image_y = image.shape[0] / 2
    center_image_x = image.shape[1] / 2
    x_coord_vec = np.linspace(0, image.shape[1], image.shape[1])
    y_coord_vec = np.linspace(0, image.shape[0], image.shape[0])
    x_coord_mat, y_coord_mat = np.meshgrid(x_coord_vec, y_coord_vec, sparse=True)
    dist_mat = np.float16(
        np.sqrt(
            ((y_coord_mat - center_image_y) ** 2)
            + ((x_coord_mat - center_image_x) ** 2)
        )
        + width / 2
    )
    gmask = np.sin(np.pi * dist_mat / (width - 1))
    gmask[dist_mat > width] = 0

    return image * gmask


def next_power_of2(number):
    return int(np.power(2, np.ceil(np.log2(number))))


def zero_pad(img, size_new):
    pad_extends = []
    dif_shape_y = size_new[0] - img.shape[0]
    dif_shape_x = size_new[1] - img.shape[1]

    pad_extends.append((dif_shape_y // 2, dif_shape_y // 2 + dif_shape_y % 2))
    pad_extends.append((dif_shape_x // 2, dif_shape_x // 2 + dif_shape_x % 2))

    padded = np.pad(img, pad_extends, "symmetric")

    return padded, pad_extends


def low_pass_filter(img, filter_width):
    """
    Low pass filter
    :param img: Image to be filtered (numpy array)
    :param filter_width: Width of the gaussian filter in fourier pixel
    :return: Return low pass filtered image
    """
    input_image_shape = img.shape
    # Zero Pad
    next_shape_y = next_power_of2(img.shape[0])
    next_shape_x = next_power_of2(img.shape[1])
    new_image_size = (next_shape_y, next_shape_x)
    padded, pad_extends = zero_pad(img, new_image_size)

    # The input image is not needed anymore at this point
    del img

    # FFT
    img_fft = fft.fft2(padded)

    # Padded image is not needed anymore
    del padded

    img_fft = fft.fftshift(img_fft)
    # Windowing with cosine window
    img_fft_windowed = window(img_fft, filter_width)
    del img_fft

    # Backward fft
    img_fft_windowed = fft.ifftshift(img_fft_windowed)
    img_filtered = fft.ifft2(img_fft_windowed)

    del img_fft_windowed
    # Crop
    img_cropped = img_filtered[
        pad_extends[0][0] : (pad_extends[0][0] + input_image_shape[0]),
        pad_extends[1][0] : (pad_extends[1][0] + input_image_shape[1]),
    ]

    del img_filtered

    return np.real(img_cropped)


def filter_images(images, filter_cutoff, filter_tmp_path, num_cpu=-1, resize_to=None):
    """
    Function will filter images gives in the image dictonary, write them into the filter_output_path

    :param images: list of images
    :param filter_cutoff: Absolute cutoff frequence (0-0.5)
    :param filter_tmp_path: Output directory for temporary files
    :return: None
    """

    if not os.path.isdir(filter_tmp_path):
        os.makedirs(filter_tmp_path)

    # Correct absolute filenames
    adapted_paths = [
        img["filename"][1:] if img["filename"][0] == "/" else img["filename"]
        for img in images
    ]

    # Create file paths for filtered files
    filtered_file_paths = [
        os.path.join(filter_tmp_path, img_pth) for img_pth in adapted_paths
    ]

    # Check if files are already existing
    already_filtered = [os.path.exists(file) for file in filtered_file_paths]

    # Get index of images that are already filtered
    index_images_already_filtered = [i for i, x in enumerate(already_filtered) if x]

    # Updates paths for already filtered images
    for index in index_images_already_filtered:
        images[index]["filename"] = filtered_file_paths[index]

    # Get index of images that have to be filtered
    index_images_to_filter = [
        i
        for i, x in enumerate(already_filtered)
        if not x and os.path.basename(filtered_file_paths[i])[0] != "."
    ]

    # Create argument tubles for them
    arg_tubles = []
    arg_tubles.extend(
        [
            (images[i]["filename"], filter_cutoff, filter_tmp_path, -1, -1, resize_to)
            for i, is_filtered in enumerate(already_filtered)
            if not is_filtered
        ]
    )

    # Do the filtering
    num_processes = None
    if num_cpu > 0:
        num_processes = num_cpu
    lock = multiprocessing.Lock()
    pool = multiprocessing.Pool(
        initializer=init, initargs=(lock,), maxtasksperchild=1, processes=num_processes
    )
    all_filtered_images_paths = pool.starmap(
        filter_single_image_and_write_to_disk, arg_tubles, chunksize=1
    )
    pool.close()
    pool.join()

    # Update paths of newly filtered images
    for i, index in enumerate(index_images_to_filter):
        images[index]["filename"] = all_filtered_images_paths[i]


def filter_single_image_and_write_to_disk(
    img_path,
    filter_cutoff,
    output_path_filtered_images,
    img_num=-1,
    total_num=-1,
    resize=None,
):
    """
    Filteres the image in img_path, save it to disk and returns the path to the filtered images
    :param img_path: Path to image to filter
    :param filter_cutoff: Absolute frequence (0 - 0.5)
    :param output_path_filtered_images: Output path for filtered images
    :param resize: tuple with image size. the image will be downsized to this size before filtering
    :return: Path to the filtered image
    """
    dirname = os.path.dirname(img_path)
    if dirname.startswith("/"):
        dirname = dirname[1:]
    filtered_image_folder = os.path.join(output_path_filtered_images, dirname)
    filtered_img_tmp_file_path = os.path.join(
        filtered_image_folder, os.path.basename(img_path)
    )
    if not os.path.exists(filtered_img_tmp_file_path):
        global LOCK
        if filter_cutoff > 0.5:
            warnings.simplefilter("once", UserWarning)
            print("Filter cutoff has to be smaller than 0.5. Set it to 0.5")
            filter_cutoff = 0.5

        if LOCK is not None:
            with LOCK:
                if not os.path.isdir(filtered_image_folder):
                    os.makedirs(filtered_image_folder)
        else:
            try:
                if not os.path.isdir(filtered_image_folder):
                    os.makedirs(filtered_image_folder)
            except FileExistsError:
                pass

        if img_num != -1 and total_num != -1:
            print(
                "Filtering ",
                img_path,
                "( Progress:",
                int(img_num / total_num * 100),
                "% )",
            )
        else:
            print("Filtering ", img_path)

        image_filtered = filter_single_image(
            img_path=img_path, filter_cutoff=filter_cutoff, resize_to=resize
        )
        imagereader.image_write(filtered_img_tmp_file_path, image_filtered)
        del image_filtered
    return filtered_img_tmp_file_path


def filter_single_image(img_path, filter_cutoff, resize_to=None):
    image = imagereader.image_read(img_path)
    cutoff_factor = 1.0
    if resize_to is not None:
        cutoff_factor = image.shape[0] / resize_to[0]
        from PIL import Image

        image = np.array(
            Image.fromarray(image).resize(resize_to, resample=Image.BILINEAR)
        )
    image_filtered = low_pass_filter(
        image, 2 * image.shape[1] * filter_cutoff * cutoff_factor
    )
    del image
    return image_filtered


def init(lock):
    """
    Init for multiprocessing
    :param lock: Lock object
    :return: None
    """
    global LOCK
    LOCK = lock



"""
Abitary Utils for crYOLO
"""
#
# COPYRIGHT
#
# All contributions by Ngoc Anh Huyn:
# Copyright (c) 2017, Ngoc Anh Huyn.
# All rights reserved.
#
# All contributions by Thorsten Wagner:
# Copyright (c) 2017 - 2019, Thorsten Wagner.
# All rights reserved.
#
# ---------------------------------------------------------------------------
#         Do not reproduce or redistribute, in whole or in part.
#      Use of this code is permitted only under licence from Max Planck Society.
#            Contact us at thorsten.wagner@mpi-dortmund.mpg.de
# ---------------------------------------------------------------------------
#
import numpy as np
import sys

from . import imagereader
import urllib.request, json

class Filament:
    """
    Filament object
    """

    def __init__(self, boxes=None):
        """
        Constructor
        :param boxes: Boxes of the filament
        """

        if boxes is None:
            self.boxes = []
        else:
            self.boxes = boxes

    def add_box(self, box):
        """
        Appends a box to the filament box list.
        :param box: box to append
        :return: filament object
        """
        self.boxes.append(box)
        return self

    def get_num_boxes(self):
        """
        :return: Number of boxes that belong the filament

        """
        return len(self.boxes)

    def does_overlap(self, box, threshold):
        """
        Compares a box with all other boxes in the filament box list and calculates the IOU overlap.
        This overlap if compared to a threshold.
        :param box: Box to compare
        :param threshold: IOU Threshold
        :return: True if IOU >= threshold
        """

        for filbox in self.boxes:
            if bbox_iou(filbox, box) >= threshold:
                return True
        return False


class BoundBox:
    """
    A bounding box of a particle
    """

    def __init__(self, x, y, w, h, c=None, classes=None):
        """
        Creates a BoundBox
        :param x: x coordinate of the center
        :param y: y coordinate of the center
        :param w: width of box
        :param h: height of the box
        :param c: confidence of the box
        :param classes: Class of the BoundBox object
        """

        self.x = x
        self.y = y
        self.w = w
        self.h = h

        self.c = c
        self.classes = classes

        self.label = -1
        self.score = -1
        self.info = None  # helping data during processing

    def get_label(self):
        """

        :return: Class with highest probability
        """
        if self.label == -1:
            self.label = np.argmax(self.classes)

        return self.label

    def get_score(self):
        """
        :return: Probability of the class
        """
        # if self.score == -1:
        self.score = self.classes[self.get_label()]

        return self.score


def getEquidistantBoxes(box1, box2, num_boxes):
    """
    Create multiple boxes between box1 and box2 with equdistant boxes
    :param box1: First box
    :param box2: Second box
    :param num_boxes: number of boxes
    :return: List of N=num_boxes equidistant boxes between box1 and box2
    """
    points = zip(
        np.linspace(box1.x, box2.x, num_boxes + 1, endpoint=True),
        np.linspace(box1.y, box2.y, num_boxes + 1, endpoint=True),
    )
    new_boxes = []
    if box1.c is None or box2.c is None:
        c = 1
    else:
        c = (box1.c + box2.c) / 2
    classes = ""
    if box1.classes:
        classes = box1.classes
    w = box1.w
    h = box2.w
    for point in points:
        b = BoundBox(x=point[0], y=point[1], w=w, h=h, c=c, classes=classes)
        new_boxes.append(b)
    return new_boxes


def resample_filaments(new_filaments, box_distance):
    """
    Resample a list of filaments
    :param new_filaments: List of filament
    :param box_distance: Target distance between two boxes
    :return: List of resampled filaments
    """
    resamples_filaments = [
        resample_filament(fil, box_distance) for fil in new_filaments
    ]
    resamples_filaments = [fil for fil in resamples_filaments if fil.boxes]
    return resamples_filaments


def resample_filament(filament, distance):
    """
    Resamples a filament. The new filament will have a specified distance between the boxes.
    :param filament: Target filament
    :param distance: Distance between two adjacent boxes.
    :return: Resampeld filament
    """
    # Find two boxes with maximum distance
    boxes = filament.boxes
    max_distance = 0
    max_boxa = None
    max_boxb = None

    for i in range(len(boxes)):
        for j in range(len(boxes)):
            if i == j:
                continue
            dist = box_squared_distance(boxes[i], boxes[j])
            if dist > max_distance:
                max_boxa = i
                max_boxb = j
    if max_boxa is not None:
        # Construct a polyline, start from max_boxa
        polyline = []
        polyline.append(boxes[max_boxa])
        assigned_boxes = []
        assigned_boxes.append(max_boxa)
        last_appended_box = boxes[max_boxa]
        last_appended_index = max_boxa
        added_new = True

        while added_new:
            min_distance = 99999999
            nearest_box_index = None
            added_new = False
            for i in range(len(boxes)):
                if i == last_appended_index:
                    continue
                if i in assigned_boxes:
                    continue

                measured_distance = box_squared_distance(last_appended_box, boxes[i])

                if measured_distance < min_distance:
                    nearest_box_index = i
                    min_distance = measured_distance

            if nearest_box_index is not None:
                polyline.append(boxes[nearest_box_index])
                assigned_boxes.append(nearest_box_index)
                last_appended_box = boxes[nearest_box_index]
                last_appended_index = nearest_box_index
                added_new = True

        # Create new boxes based on the polyline
        new_boxes_candidates = []
        sqdistance = distance * distance

        for i in range(len(polyline) - 1):
            eq_boxes = getEquidistantBoxes(polyline[i], polyline[i + 1], num_boxes=100)
            new_boxes_candidates.extend(eq_boxes)
        new_boxes = []
        dist = -1
        for box in new_boxes_candidates:
            if dist == -1:
                new_boxes.append(box)
                dist = 0
            else:
                dist = box_squared_distance(new_boxes[len(new_boxes) - 1], box)
                if dist > sqdistance:
                    new_boxes.append(box)

        new_filament = Filament()
        for box in new_boxes:
            new_filament.add_box(box)

        return new_filament
    return None


def box_squared_distance(boxa, boxb):
    """
    calculates the squared distance between the center of two boxes.
    :param boxa: First box
    :param boxb: Second box
    :return: Squared distance between boxa and boxb
    """
    return np.square(boxa.x - boxb.x) + np.square(boxa.y - boxb.y)



def normalize(image, margin_size=0):
    """
    Normalize an image. If wanted, a specified relative margin can be ignored.
    :param image: Image to normalize
    :param margin_size: Relative size of the margin to be ignored during normalization. Number between 0 - 1.
    :return: Normalized image
    """
    if margin_size < 0 or margin_size > 1:
        print("Normalization has to be between 0 and 1.")
        if(margin_size<0):
            margin_size = 0
        else:
            margin_size = 1
        print("Set it to", margin_size)

    if not np.issubdtype(image.dtype, np.float32):
        image = image.astype(np.float32)
    mask = np.s_[
        int(image.shape[0] * (margin_size)) : int(image.shape[0] * (1 - margin_size)),
        int(image.shape[1] * margin_size) : int(image.shape[1] * (1 - margin_size)),
    ]
    img_mean = np.mean(image[mask])
    img_std = np.std(image[mask])

    image = (image - img_mean) / (3 * 2 * img_std)

    return image


def bbox_iou(box1, box2):
    """
    Calculates the intersection of union for two boxes.
    :param box1: First box
    :param box2: Second box
    :return: Intersection of union
    """
    x1_min = box1.x - box1.w / 2
    x1_max = box1.x + box1.w / 2
    y1_min = box1.y - box1.h / 2
    y1_max = box1.y + box1.h / 2

    x2_min = box2.x - box2.w / 2
    x2_max = box2.x + box2.w / 2
    y2_min = box2.y - box2.h / 2
    y2_max = box2.y + box2.h / 2

    intersect_w = interval_overlap([x1_min, x1_max], [x2_min, x2_max])
    intersect_h = interval_overlap([y1_min, y1_max], [y2_min, y2_max])

    intersect = intersect_w * intersect_h

    union = box1.w * box1.h + box2.w * box2.h - intersect

    return float(intersect) / union


def interval_overlap(interval_a, interval_b):
    """
    Calculates the overlap between two intervals
    :param interval_a: Tuple with two elements (lower and upper bound)
    :param interval_b: Tuple with two elements (lower and upper bound)
    :return: Overlap between two intervals
    """
    x1, x2 = interval_a
    x3, x4 = interval_b

    if x3 < x1:
        if x4 < x1:
            return 0
        else:
            return min(x2, x4) - x1
    else:
        if x2 < x3:
            return 0
        else:
            return min(x2, x4) - x3


def sigmoid(x):
    """
    Sigmoid function
    :param x: argument
    :return: sigmoid of x
    """
    return 1.0 / (1.0 + np.exp(-x))


def filter_image_noise2noise(img_path, model_path, padding=15, batch_size=4):
    """
    Filters an images using JANNI (noise2noise implementation)
    :param img_path: Input image
    :param model_path: Path to JANNI model
    :param padding: Padding to reduce edge effects
    :param batch_size: How many batches used in denoising
    :return: Denoised image
    """
    import h5py
    from janni import predict as janni_predict
    from janni import models as janni_models

    with h5py.File(model_path, mode="r") as f:
        try:
            import numpy as np

            model = str(np.array((f["model_name"])))
            patch_size = tuple(f["patch_size"])
        except KeyError:
            print("Not supported filtering model. Stop.")
            sys.exit(0)

    if model == "unet":
        model = janni_models.get_model_unet(input_size=patch_size)
        model.load_weights(model_path)
    else:
        print("Not supported model", model, "Stop")
        sys.exit(0)

    image = imagereader.image_read(img_path)
    fitlered_img = janni_predict.predict_np(
        model, image, patch_size=(1024, 1024), padding=padding, batch_size=batch_size
    )

    return fitlered_img


def filter_images_noise2noise_dir(
    img_paths,
    output_dir_filtered_imgs,
    model_path,
    padding=15,
    batch_size=4,
    resize_to=None,
):
    """
    Filters multiple images using JANNI (noise2noise implementation) and save it to disk
    :param img_paths: List of images paths
    :param output_dir_filtered_imgs: Path to write denoised images
    :param model_path: Path to JANNI model
    :param padding: Padding to reduce edge effects
    :param batch_size: How many batches used in denoising
    :param resize_to: Size to which to output is resized
    :return:
    """
    import h5py
    from janni import predict as janni_predict
    from janni import models as janni_models

    with h5py.File(model_path, mode="r") as f:
        try:
            import numpy as np

            model = str(np.array((f["model_name"])))
            patch_size = tuple(f["patch_size"])
        except KeyError:
            print("Not supported filtering model. Stop.")
            sys.exit(0)

    if model == "unet":
        model = janni_models.get_model_unet(input_size=patch_size)
        model.load_weights(model_path)
    else:
        print("Not supported model", model, "Stop")
        sys.exit(0)

    filtered_paths = janni_predict.predict_list(
        image_paths=img_paths,
        output_path=output_dir_filtered_imgs,
        model=model,
        patch_size=patch_size,
        padding=padding,
        batch_size=batch_size,
        output_resize_to=resize_to,
    )

    return filtered_paths

def check_for_updates():
    """
    Checks if cryolo updates are available
    :return: None
    """
    try:
        import cryolo
        with urllib.request.urlopen(
                "https://pypi.org/pypi/cryolo/json") as url:
            data = json.loads(url.read().decode())
            from distutils.version import LooseVersion as V
            current_version = cryolo.__version__
            latest_version = current_version
            for ver in data["releases"].keys():
                vers = V(ver)
                if vers > V(latest_version) and len(vers.version) == 3:
                    latest_version = ver
            if V(latest_version) > V(current_version):
                print("###############################################")
                print("New version of crYOLO available")
                print("Local version:\t\t", current_version)
                print("Latest version:\t\t", latest_version)
                print("More information here:\n",
                      "http://sphire.mpg.de/wiki/doku.php?id=pipeline:window:cryolo")
                print("###############################################")
    except Exception as e:
        print(e)
        pass


#
# COPYRIGHT
#
# All contributions by Thorsten Wagner:
# Copyright (c) 2017 - 2019, Thorsten Wagner.
# All rights reserved.
#
# ---------------------------------------------------------------------------
#         Do not reproduce or redistribute, in whole or in part.
#      Use of this code is permitted only under licence from Max Planck Society.
#            Contact us at thorsten.wagner@mpi-dortmund.mpg.de
# ---------------------------------------------------------------------------
#

from keras.callbacks import ModelCheckpoint
import h5py


class MyModelCheckpoint(ModelCheckpoint):
    def __init__(
        self,
        filepath,
        monitor="val_loss",
        verbose=0,
        save_best_only=False,
        save_weights_only=False,
        mode="auto",
        period=1,
    ):
        pass

    def on_epoch_end(self, epoch, logs=None):
        current = logs.get(self.monitor)
        model_improved = self.monitor_op(current, self.best)
        super(ModelCheckpoint, self).on_epoch_end(epoch, logs)
        if model_improved:

            with h5py.File(self.filepath, mode="r+") as f:
                f["anchors"] = self.anchors
